function [ x, out, opts ] = tfocs_N83( smoothF, affineF, projectorF, x0, opts )
% TFOCS_N83 Nesterov's 1983 accelerated method; also by Beck and Teboulle 2005 (FISTA).
% [ x, out, opts ] = tfocs_N83( smoothF, affineF, nonsmoothF, x0, opts )
%   Implements Nesterov's 1983 method.
%   A variety of calling sequences are supported; type 
%      help tfocs_help
%   for a full explanation.

% Initialization
alg = 'N83';
algorithm = 'Nesterov''s 1983 single-projection method';
alpha = 0; beta = 0; mu = 0; L = 0; % Do not remove: necessary because of a MATLAB quirk
if(~isfield(opts, 'max_time'))
    opts.max_time = inf;
end
tbound = opts.max_time;
tfocs_initialize
if nargin == 0, return; end
tic
outtimes = 0;
[n, k] = size(x);
temp = z;
clear('z');
z.U = temp;
z.D = eye(size(z.U, 2));
nf = 0;
ng = 0;
while true,
      x_old =   x;
      z_old =   z;
    
    % The backtracking loop
    L_old      = L;
   	L          = L * alpha;
    theta_old  = theta;
    times = 0;
    while true,
        times = times + 1;
        % Acceleration
        theta = advance_theta( theta_old, L, L_old );
        % Next iterate
        if theta < 1,
            y = low_rank_addition(( 1 - theta ), x_old , theta, z_old, k);
            f_y = Inf;
            g_y = [];
        end
        % Compute function values
        if isempty( g_y ),
            [ f_y, g_y ] = apply_smooth( y ); 
            nf = nf + 1;
            ng = ng + 1;
        end

        % Standard gradient
        x = low_rank_add_gfv(1, y, - (1/L), g_y, k);
        f_x = Inf;
        g_x = [];
        % TFOCS_BACKTRACK
        % Quick exit for no backtracking
        if beta >= 1, break; end % SRB changing == to >=
        % Quick exit if no progress made
        xy = low_rank_substraction(1, x, -1, y);
        temp = xy.U' * xy.U;
        xy_sq = real(trace(xy.D * temp * xy.D' * temp));
        if xy_sq == 0, localL = Inf; 
            break; 
        end

        % Compute Lipschitz estimate
        if backtrack_simple,
            if isinf( f_x ),
                f_x = apply_smooth( x );
                nf = nf + 1;
            end
            first_order = inpro_gf_v(g_y, xy);
            q_x = f_y + first_order + 0.5 * L * xy_sq;

            localL = L + 2 * max( f_x - q_x, 0 ) / xy_sq;
            backtrack_simple = abs( f_y - f_x ) >= backtrack_tol * max( abs( f_x ), abs( f_y ) );
%             fprintf('\t f_x - f_y:%e, inpro:%e, second_order:%e\n', f_x - f_y, first_order, 0.5 * L * xy_sq);%%---
%             fprintf('\t L:%e, localL:%e, f_x:%e, q_x:%e, f_y:%e\n', L, localL, f_x, q_x, f_y);%%---
%             fprintf('\t diff_f:%e, fx:%e, fy:%e, backtrack_simple:%d, backtrack_tol:%e\n', abs( f_y - f_x ), f_x, f_y, backtrack_simple, backtrack_tol);%%---
        else
            if isempty( g_x ),
                [ f_x, g_x ] = apply_smooth( x );
                nf = nf + 1;
                ng = ng + 1;
            end
            localL = 2 * inpro_gf1addgf2_v(1, g_x, -1, g_y, xy) / xy_sq;
        end

        % Exit if Lipschitz criterion satisfied, or if we hit Lexact
        backtrack_steps = backtrack_steps + 1;
        if localL <= L || L >= Lexact,
            break; 
        end
        
        if ~isinf( localL ), 
            L = min( Lexact, localL );
        elseif isinf( localL ), localL = L; end
        L = min( Lexact, max( localL, L / beta ) );
        % TFOCS v1.3 by Stephen Becker, Emmanuel Candes, and Michael Grant.
        % Copyright 2013 California Institute of Technology and CVX Research.
        % See the file LICENSE for full license information.

    end
%     fprintf('out-iter:%d, in-iter:%d, f:%e \n', outtimes, times, f_x);
    outtimes = outtimes + 1;

    % Collect data, evaluate stopping criteria, and print status
%SOLVER_ITERATE    TFOCS helper script
%	Performs the iterate processing common to all of the first-order solvers
%
%   Major inputs: x, x_old, xy , A_x, A_y, g_Ax, g_Ay 
%       (does not need g_x or g_y really)

% Test for positive stopping criteria
n_iter = n_iter + 1;
norm_x = sqrt( tfocs_normsq( x' * x ) );
xmxold = low_rank_substraction(1, x, -1, x_old);
temp = xmxold.U' * xmxold.U;
norm_dx = sqrt(real(trace(xmxold.D * temp * xmxold.D' * temp)));
% norm_dx = sqrt( tfocs_normsq( xmxold' * xmxold ) );
% We also handle legacy stopping criteria used in the paper:
if stopCrit == 2 && beta >= 1
    % xy_sq not already computed
    xy      = low_rank_substraction(1, x, -1, y);
    temp = xy.U' * xy.U;
    xy_sq = real(trace(xy.D * temp * xy.D' * temp));
%     xy_sq   = tfocs_normsq( xy' * xy );
end
if isnan( f_y ),
	status = 'NaN found -- aborting';
elseif (stopCrit == 1) && norm_dx  == 0
    if n_iter > 1
        status = 'Step size tolerance reached (||dx||=0)';
    end
elseif (stopCrit == 1) && norm_dx < tol * max( norm_x, 1 ),
    status = sprintf('Step size tolerance reached, norm_x:%e', norm_x);
elseif (stopCrit == 1) && toc >= tbound
    status = sprintf('Computational time reached, time:%f', tbound);
elseif (stopCrit == 2) && L*sqrt(xy_sq) < tol,
    status = 'Step size tolerance reached';
elseif n_iter == maxIts,
	status = 'Iteration limit reached';
elseif countOps && any( tfocs_count___ >= maxCounts ),
	status = 'Function/operator count limit reached';
elseif backtrack_steps > 0 && xy_sq == 0,
    status = 'Unexpectedly small stepsize';
end

% for stopCrit 3, we need the new and old dual points
if stopCrit == 3 || stopCrit == 4
    if ~saddle, error('stopCrit = {3,4} requires a saddle point problem'); end
    if exist('cur_dual','var')
        old_dual = cur_dual;
    else
        old_dual = [];
    end
end

%
% For stopping criteria or algorithm control, we assume that the user
% needs the objective function value, but does not wish to do any more
% computation than necessary. So we will use the function value for y
% instead of x if that is cheaper to obtain. So here we determine what
% the additional computational costs will be, and choose the path that
% minimizes them, favoring x in the case of a tie.
%

% if isempty(status) && ( ~isempty(stopFcn) || restart < 0 || stopCrit == 3 || stopCrit ==4 ),
if (isempty(status) || ~isempty(findstr(status,'limit')) ) ...
        && ( ~isempty(stopFcn) || restart < 0 || stopCrit == 3 || stopCrit ==4 ),
    comp_x = [ isinf(f_x), saddle*~isempty(stopFcn)*isempty(g_Ax), isinf(C_x) ];
    comp_y = [ isinf(f_y), saddle*~isempty(stopFcn)*isempty(g_Ay), isinf(C_y) ];
    if sum(comp_x) <= sum(comp_y),
        if comp_x(2), [f_x,g_Ax] = apply_smooth(A_x);
        elseif comp_x(1), f_x = apply_smooth(A_x); end
        if comp_x(3), C_x = apply_projector(x); end
        cur_pri = x; cur_dual = g_Ax;
        f_v = maxmin*(f_x+C_x);
    else
        if comp_y(2), [f_y,g_Ay] = apply_smooth(A_y);
        elseif comp_y(1), f_y = apply_smooth(A_y); end
        if comp_y(3), C_y = apply_projector(y); end
        cur_pri = y; cur_dual = g_Ay;
        f_v = maxmin*(f_y+C_y);
    end
    for err_j = 1 : numel(stopFcn),
        if saddle,
            stop = stopFcn{err_j}(f_v,cur_pri, get_dual(cur_dual) );
        else
            stop = stopFcn{err_j}(f_v,cur_pri);
        end
        if stop
            status = sprintf('Reached user''s supplied stopping criteria no. %d',err_j);
        end
    end
end

% Now we can apply stopCrit 3 if it has been requested:
if (stopCrit == 3 || stopCrit == 4 ) 
    if ~isempty( old_dual ) && ~isempty( cur_dual )
        d_dual = sqrt(tfocs_normsq( old_dual - cur_dual ));
        if isnumeric( cur_dual ),
             norm_cur = tfocs_normsq( cur_dual );
             norm_old = tfocs_normsq( old_dual );
        else
            norm_cur = 0;
            norm_old = 0;
            for j = 1:max( [1, numel(cur_dual)-1] )
                norm_cur = norm_cur  + tfocs_normsq( cur_dual{j}  );
                norm_old = norm_old  + tfocs_normsq( old_dual{j}  );
            end
        end
        norm_cur = sqrt(norm_cur);
        norm_old = sqrt(norm_old);
    else
        d_dual   = Inf;
        norm_cur = 0; 
        norm_old = 0;
    end
    
	% Note: it is common for the duals to be stuck at zero for quite a few
    %   iterations at the beginning.  In stopCrit = 4 mode, we will not
    %   terminate if both the current and old dual are zero.
    
    if stopCrit == 4
        % for "4", we look at relative change, not absolute change
        if norm_cur > 10*eps && norm_old > 10*eps
            d_dual = d_dual / norm_cur;
        else
            % The dual vectors are zero, so do not terminate.
            % This is equivalent to defining 0/0 = Inf;
            d_dual = Inf;
        end
    end
    if d_dual < tol
        status = 'Step size tolerance reached';
    end
end



%
% Data collection. Since this is used only for algorithm analysis and not
% for algorithm control, we are free to make additional computations here
% without adding them to our total algorithm cost. We prefer to track the
% x sequence in this case, since it is what we will ultimately choose as
% the solution at the end of the algorithm.
%

will_print = fid && printEvery && ( ~isempty( status ) || ~mod( n_iter, printEvery ) );
if saveHist || will_print,
    f_x_save = f_x;
    g_Ax_save = g_Ax;
    if ~isempty(errFcn) && saddle,
        if isempty(g_Ax),
            [ f_x, g_Ax ] = smoothF( A_x );
        end
        out.dual = get_dual( g_Ax );
    end
    if isinf(f_x),
        f_x = smoothF( A_x );
    end
    if isinf( C_x ),
        C_x = projectorF( x ); 
    end
    f_w = maxmin * ( f_x + C_x );
    if ~isempty(errFcn) && iscell(errFcn)
        for err_j = 1 : numel(errFcn),
            if saddle,
                errs(err_j) = errFcn{err_j}(f_w,x,out.dual);
            else
                errs(err_j) = errFcn{err_j}(f_w,x);
            end
        end
    end
    f_x = f_x_save;
    g_Ax = g_Ax_save;
end

% Register a warning if the step size suggests a Lipschitz violation
if isempty(status) && ( beta < 1 && backtrack_simple && localL > Lexact ),
    warning_lipschitz = true;
end	

% Print status
if will_print,
    if warning_lipschitz,
        warning_lipschitz = false;
        bchar = 'L';
    elseif backtrack_simple,
        bchar = ' '; 
    else
        bchar = '*'; 
    end
	fprintf( fid, '%-4d| %+12.5e  %8.2e  %8.2e%c', ...
        n_iter, f_w, norm_dx / max( norm_x, 1 ), 1 / L, bchar );
    if countOps,
        fprintf( fid, '|' );
        fprintf( fid, ' %5d', tfocs_count___ );
    end
    if ~isempty(errFcn),
        if countOps,
            fprintf( fid, ' ' );
        end
    	fprintf( fid, '|' );
        fprintf( fid, ' %8.2e', errs );
    end
    
    
    if printStopCrit
        % Display the number used to determine stopping

        switch stopCrit
            case 1
                if exist('norm_dx','var') && exist('norm_x','var')
                    stopResid   =  norm_dx/max( norm_x,1);
                else
                    stopResid   = Inf;
                end
            case 2
                stopResid   = L*sqrt(xy_sq);
            case {3,4}
                if exist('d_dual','var')
                    stopResid   = d_dual;
                end
            case Inf
                % do nothing
                stopResid = 0;
            otherwise
                error('Bad stopCrit value');
        end
    
        if ~isempty(errFcn) || countOps
            fprintf( fid, ' ' );
        end
        fprintf( fid, '|' );
        fprintf( fid, ' %8.2e', stopResid );
    end
    
	fprintf( fid, '\n');
end

% Save history, extending arrays if needed
if saveHist,
    if length(out.f) < n_iter && isempty(status),
        csize = min(maxIts,length(out.f)+1000);
        out.f(end+1:csize,1) = 0;
        out.theta(end+1:csize,1) = 0;
        out.stepsize(end+1:csize,1) = 0;
        out.normGrad(end+1:csize,1) = 0;
        if countOps,
            out.counts(end+1:csize,:) = 0;
        end
        if ~isempty(errs),
            out.err(end+1:csize,:) = 0;
        end
    end
    out.f(n_iter) = f_w;
    out.theta(n_iter) = theta;
    out.stepsize(n_iter) = 1 / L;
    out.normGrad(n_iter) = norm_dx;
    if countOps,
        out.counts(n_iter,:) = tfocs_count___;
    end
    if ~isempty(errFcn),
        out.err(n_iter,:) = errs;
    end
end

% Exit if positive stopping criteria met
if ~isempty( status ),
	break;
end

% Restart acceleration if necessary
backtrack_steps = 0;
% "No regress" feature: test was (maxmin*f_v > f_v_old)
%   This worked for minimization, but not for maximization! Dec 15 2010.
% Fixed it.
% two changes: (1) test is now ( maxmin*f_v > maxmin*f_v_old)
%              (2) to reset f_v_old, set to maxmin*Inf, not just +Inf
if n_iter - restart_iter == abs(round(restart)) || (restart < 0 && maxmin*f_v > maxmin*f_v_old),
    restart_iter = n_iter;
    backtrack_simple = true;
	theta = Inf;
    y = x; A_y = A_x; f_y = f_x; g_Ay = g_Ax; g_y = g_x; C_y = C_x;
    z = x; A_z = A_x; f_z = f_x; g_Az = g_Ax; g_z = g_x; C_z = C_x;
    f_v_old = maxmin*Inf; % important!
%     continue;
elseif restart < 0,
    f_v_old = f_v;
end

% TFOCS v1.3 by Stephen Becker, Emmanuel Candes, and Michael Grant.
% Copyright 2013 California Institute of Technology and CVX Research.
% See the file LICENSE for full license information.


    % Reversed z update. This step is skipped if restart occurs
    if theta == 1,
        z.U   =   x;
        z.D = eye(size(z.U, 2));
    else
        z = low_rank_substraction((1/theta), x , ( 1 - 1/theta ), x_old);
    end    
end
t = toc;
fprintf('timecost:%f \n', t);
opts.timecost = t;
opts.iter = outtimes;
opts.nf = nf;
opts.ng = ng;
opts.f_final = f_x;

% Final processing
tfocs_cleanup

% TFOCS v1.3 by Stephen Becker, Emmanuel Candes, and Michael Grant.
% Copyright 2013 California Institute of Technology and CVX Research.
% See the file LICENSE for full license information.

function output = low_rank_addition(a, x, b, y, k)
    if(a < 0 || b < 0)
        error('a and b must be nonnegative!');
    end
    [n, k1] = size(x);
    k2 = size(y, 2);
    Z = [x, y.U];
    [Q, R] = qr(Z, 0);
    M = R * diag([ones(k1, 1) * a; diag(y.D) * b]) * R';
    [U, D] = eig(M);
    D = real(D);
    [d, idx] = sort(diag(D), 'descend');
    output = Q * U(:, idx(1:k)) * diag(sqrt(max(d(1:k), 0)));
return


function output = low_rank_substraction(a, x, b, y)
    if(a < 0 || b > 0)
        error('a must be nonnegative and b must be nonpositive');
    end
    [n, k1] = size(x);
    k2 = size(y, 2);
    Z = [x, y];
    [Q, R] = qr(Z, 0);
    M = R * diag([ones(k1, 1) * a; ones(k2, 1) * b]) * R';
    [U, D] = eig(M);
    D = real(D);
    output.U = Q * U;
    output.D = D;
return

function output = gf2norm(gf, n, k)
    func = @(v)gf(v);
    opts.isreal = false;
    opts.issym = 1;
    [U, D] = eigs(func, n, 1, 'lm', opts);
    output = abs(D);
return


function output = low_rank_add_gfv(a, x, b, gf, k)
    n = size(x, 1);
    func = @(v) a * x * (x' * v) + b * gf(v);
    opts.isreal = false;
    opts.issym = 1;
    [U, D] = eigs(func, n, k, 'lr', opts);
    D = real(D);
    D = max(D, 0);
    output = U * sqrt(D);
return

function output = inpro_gf_v(gf, eta)
    temp = gf(eta.U) * eta.D';
    output = real(eta.U(:)' * temp(:));
return

function output = inpro_gf1addgf2_v(a, gf1, b, gf2, v)
    temp = (a * gf1(v.U) + b * gf2(v.U)) * v.D;
    output = real(v.U(:)' * temp(:));
return
